﻿#include "shm_object.h"
#include "shm_manager.h"
#include <inttypes.h>
#if defined(__GNUC__)
#include <sys/types.h>
#include <dirent.h>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/stat.h> /* For mode constants */
#include <fcntl.h>    /* For O_* constants */
#include <unistd.h>
#define SHM_PATH "/dev/shm"
namespace bg
{
    namespace detail
    {
        bool GetDirectorySize(const char *path, size_t *const total_size)
        {
            char full_path[264]; // [rsp+A0h] [rbp-130h] BYREF
            struct stat st;
            DIR *path_dir = opendir(path);
            if (!path_dir)
            {
                SHM_ERROR("opendir() failed, path: %s, error: %s.", path, strerror(*__errno_location()));
                return false;
            }
            struct dirent64 *diren = readdir64(path_dir);
            if (!diren)
            {
                closedir(path_dir);
                return true;
            }
            while (1)
            {
                snprintf(full_path, 264, "%s/%s", path, diren->d_name);

                if (diren->d_type != DT_DIR)
                    break;

                if (strcmp(diren->d_name, ".") && strcmp(diren->d_name, ".."))
                {
                    st.st_dev = 0LL;
                    if (!bg::detail::GetDirectorySize(full_path, &st.st_dev))
                    {
                        closedir(path_dir);
                        return false;
                    }
                    *total_size += st.st_dev;
                    SHM_DEBUG("directory entry %s, size: %#" PRIx64 "", full_path, st.st_dev);
                }
            READ_DIR:
                diren = readdir64(path_dir);
                if (!diren)
                {
                    closedir(path_dir);
                    return true;
                }
            }
            if (diren->d_type != DT_REG)
            {
                SHM_WARN("entry %s, type: %u, ignored.", full_path, diren->d_type);
                goto READ_DIR;
            }
            if (!__lxstat64(1, full_path, (struct stat64 *)&st))
            {
                *total_size += st.st_size;
                SHM_TRACE("file entry %s, size: %#" PRIx64 "", full_path, st.st_size);
                goto READ_DIR;
            }

            SHM_ERROR("lstat() failed, file: %s, error : %s.", full_path, strerror(*__errno_location()));

            closedir(path_dir);
            return false;
        }

        void *ShmObjectCreate(const char *path, uintptr_t fixed_addr, size_t *real_size, size_t *mmap_size)
        {
            size_t page_size = sysconf(_SC_PAGESIZE);
            int fd = shm_open(path, O_CREAT | O_RDWR, 0666);
            if (fd == -1)
            {
                int *error = __errno_location();

                if (*error != 17)
                {
                    SHM_ERROR("shm_open() failed, path: %s, error: %s.", path, strerror(*__errno_location()));
                    return nullptr;
                }

                SHM_INFO("shm object(%s) already exists, will be deleted.", path);
                if (shm_unlink(path))
                {
                    SHM_ERROR("shm_unlink() failed, path: %s, error: %s.", path, strerror(*__errno_location()));
                }
                else
                {
                    fd = shm_open(path, O_CREAT | O_RDWR, 0666);
                    if (SHM_UNLIKELY(fd == -1))
                    {
                        SHM_ERROR("shm_open() failed, path: %s, error: %s.", path, strerror(*__errno_location()));
                        return nullptr;
                    }
                }
            }

            *real_size = PALIGN_UP(*real_size, page_size);
            size_t available_size;

            if (!bg::detail::GetAvailableMemorySize(&available_size) || available_size < *real_size)
            {
                SHM_ERROR("shared memory not enough, available: %#" PRIx64 ", required: %#" PRIx64 ".", available_size, *real_size);
                close(fd);
                shm_unlink(path);
                return nullptr;
            }

            if (SHM_UNLIKELY(ftruncate64(fd, *real_size)))
            {
                SHM_ERROR("ftruncate() failed, path: %s, size : %#" PRIx64 ", error: %s.", path, *real_size, strerror(*__errno_location()));
                close(fd);
                shm_unlink(path);
                return nullptr;
            }
            int flag;
            if (*mmap_size < *real_size)
            {
                SHM_INFO("mmap size(%#" PRIx64 ") is not enough, revised to real size(%#" PRIx64 ").", *mmap_size, *real_size);
                *mmap_size = *real_size;
            }
            *mmap_size = PALIGN_UP(*mmap_size, page_size);

            if (fixed_addr)
            {
                if (fixed_addr & (page_size - 1))
                {
                    uintptr_t new_addr = PALIGN_UP(fixed_addr, page_size);
                    SHM_INFO("ixed address(%#" PRIx64 ") is not aligned to page size(%#" PRIx64 "), revised to %#" PRIx64 ".", fixed_addr, page_size, new_addr);
                    fixed_addr = new_addr;
                }
                flag = MAP_SHARED;
            }
            else
            {
                flag = MAP_FIXED;
            }

            void *mem = mmap64((void *)fixed_addr, *mmap_size, PROT_READ | PROT_WRITE, flag, fd, 0);
            if (mem == MAP_FAILED)
            {
                SHM_ERROR("mmap() failed, addr(%#" PRIx64 "), size(%#" PRIx64 "), error: %s.", fixed_addr, *mmap_size, strerror(*__errno_location()));
                close(fd);
                shm_unlink(path);
                return nullptr;
            }

            close(fd);
            return mem;
        }

        void *ShmObjectAttach(const char *path, uintptr_t fixed_addr, size_t *real_size, size_t *mmap_size)
        {
            size_t page_size = sysconf(_SC_PAGESIZE);
            struct stat st;
            int fd = shm_open(path, O_RDWR, 0666);
            if (SHM_UNLIKELY(fd == -1))
            {
                SHM_ERROR("shm_open() failed, path: %s, error: %s.", path, strerror(*__errno_location()));
                close(fd);
                return nullptr;
            }

            if (__fxstat64(1, fd, (struct stat64 *)&st))
            {
                SHM_ERROR("fstat() failed, path: %s, error: %s.", path, strerror(*__errno_location()));

                close(fd);
                return nullptr;
            }

            *real_size = st.st_size;
            if (static_cast<size_t>(st.st_size) > *mmap_size)
            {
                SHM_INFO("mmap size(%#" PRIx64 ") is not enough, revised to real size(%#" PRIx64 ").", *mmap_size, *real_size);
                *mmap_size = *real_size;
            }

            *mmap_size = PALIGN_UP(*mmap_size, page_size);
            int flag;
            if (fixed_addr)
            {
                if (fixed_addr & (page_size - 1))
                {
                    uintptr_t new_addr = PALIGN_UP(fixed_addr, page_size);
                    SHM_INFO("fixed address(%#" PRIx64 ") is not aligned to page size(%#" PRIx64 "), revised to %#" PRIx64 ".", fixed_addr, page_size, new_addr);
                }

                flag = MAP_SHARED;
            }
            else
            {
                flag = MAP_FIXED;
            }

            void *memory = mmap64((void *)fixed_addr, *mmap_size, PROT_READ | PROT_WRITE, flag, fd, 0);
            if (memory != (void *)-1)
            {
                close(fd);
                return memory;
            }

            SHM_ERROR("mmap() failed, addr(%p), size(%#" PRIx64 "), error: %s.", memory, *mmap_size, strerror(*__errno_location()));
            close(fd);
            return nullptr;
        }

        void ShmObjectDelete(void *addr, uintptr_t mmap_size, const char *path)
        {
            if (munmap(addr, mmap_size))
            {
                SHM_ERROR("munmap() failed, addr(%p), mmap size(%#" PRIx64 "), error: %s.", addr, mmap_size, strerror(*__errno_location()));
            }
            if (path)
            {
                if (shm_unlink(path))
                {
                    SHM_ERROR("shm_unlink() failed, path: %s, error: %s.", path, strerror(*__errno_location()));
                }
            }
        }

        bool ShmObjectResize(const char *path, size_t *new_size)
        {
            size_t available_size = 0;
            struct stat st;
            size_t page_size = sysconf(_SC_PAGESIZE);
            int fd = shm_open(path, O_RDWR, 0666);
            size_t diff = 0;
            if (SHM_UNLIKELY(fd == -1))
            {
                SHM_ERROR("shm_open() failed, path: %s, error : %s.", path, strerror(*__errno_location()));
                return false;
            }
            if (__fxstat64(1, fd, (struct stat64 *)&st))
            {
                SHM_ERROR("fstat() failed, path: %s, error: %s.", path, strerror(*__errno_location()));
                close(fd);
                return false;
            }

            *new_size = PALIGN_UP(*new_size, page_size);
            if (SHM_UNLIKELY(!bg::detail::GetAvailableMemorySize(&available_size)))
            {
                close(fd);
                return false;
            }
            if (static_cast<size_t>(st.st_size) < *new_size)
            {
                diff = *new_size - st.st_size;

                if (diff > available_size)
                {
                    SHM_ERROR("shared memory not enough, available: %#" PRIx64 ", required: %#" PRIx64 ".", available_size, diff);
                    close(fd);
                    return false;
                }
            }

            if (!ftruncate64(fd, *new_size))
            {
                close(fd);
                return true;
            }

            SHM_ERROR("ftruncate() failed, path: %s, size: %#" PRIx64 ", error: %s.", path, *new_size, strerror(*__errno_location()));
            return false;
        }

        bool GetAvailableMemorySize(size_t *const size)
        {
            struct statvfs st; // [rsp+10h] [rbp-80h] BYREF

            if (statvfs64("/dev/shm", (struct statvfs64 *)&st))
            {
                SHM_ERROR("statvfs() failed, path: %s, error: %s", SHM_PATH, strerror(*__errno_location()));
                return false;
            }
            size_t used_size = 0LL;
            bool ret = bg::detail::GetDirectorySize(SHM_PATH, &used_size);
            if (SHM_UNLIKELY(!ret))
            {
                return false;
            }

            __fsblkcnt64_t total_size = st.f_blocks * st.f_bsize;
            double big_number = total_size * 95;
            big_number /= 100;
            uint64_t hwm_size = static_cast<uint64_t>(big_number);        
            
            if (total_size < used_size)
            {
                return false;
            }
            *size = hwm_size - used_size;
            SHM_DEBUG("%s, used size: %#lx, total size %#lx(%#lx, %#lx), hwm size: %#lx", SHM_PATH, used_size, total_size, st.f_blocks, st.f_blocks, hwm_size);
            return true;
        }

    }
}
#else

#include <windows.h>
#include <Psapi.h>

static std::unordered_map<void *, HANDLE> g_mapping_handles;

namespace bg
{
    namespace detail
    {
        bool GetDirectorySize(const char *path, size_t *const total_size)
        {
            HANDLE handle = GetCurrentProcess();
            if (handle == nullptr)
            {
                return false;
            }
            PROCESS_MEMORY_COUNTERS pmc;
            GetProcessMemoryInfo(handle, &pmc, sizeof(pmc));
            *total_size = pmc.WorkingSetSize;
            return true;
        }

        void *ShmObjectCreate(const char *path, uintptr_t fixed_addr, size_t *real_size, size_t *mmap_size)
        {
            SYSTEM_INFO si;
            GetSystemInfo(&si);
            size_t pagesize = si.dwPageSize;

            size_t align = pagesize - 1;

            *real_size = PALIGN_UP(*real_size, pagesize);
            *mmap_size = PALIGN_UP(*mmap_size, pagesize);

            // windows ��û���ҵ���̬��չ�����ڴ�취 ����һ��ʼ����������
            if (*mmap_size != *real_size)
            {
                SHM_WARN("real size(%#llx) is not enough, revised to mmap size(%#llx).", *real_size, *mmap_size);
                *real_size = *mmap_size;
            }

            HANDLE handle = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE,
                                              (uint32_t)(*real_size >> 32),
                                              (uint32_t)(*real_size & 0xffffffff),
                                              path);
            void *addr = nullptr;
            if (handle == nullptr)
            {
                SHM_ERROR("CreateFileMapping(%zu, %s) failed error:%s", *real_size, path, strerror(GetLastError()));
                return nullptr;
            }

            if (GetLastError() == ERROR_ALREADY_EXISTS)
            {
                SHM_WARN("shm object(%s) already exists, will be deleted.", path);
                if (CloseHandle(handle))
                {
                    SHM_ERROR("CloseHandle failed, path: %s, error: %s.", path, strerror(GetLastError()));
                }

                handle = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE,
                                           (uint32_t)(*real_size >> 32),
                                           (uint32_t)(*real_size & 0xffffffff),
                                           path);

                if (handle == nullptr)
                {
                    SHM_ERROR("CreateFileMapping failed, path: %s, error: %s.", strerror(GetLastError()));
                    return nullptr;
                }

                if (GetLastError() != NO_ERROR)
                {
                    goto END;
                }
            }

            if (fixed_addr)
            {
                uintptr_t base_addr = fixed_addr;
                if (base_addr & align)
                {
                    base_addr = PALIGN_UP(fixed_addr, pagesize);
                    SHM_WARN("fixed address(%#" PRIx64 ") is not aligned to page size(%#" PRIx64 "), revised to %#" PRIx64 ".", base_addr, pagesize, fixed_addr);
                    fixed_addr = base_addr;
                }
            }

            addr = MapViewOfFileEx(handle, FILE_MAP_WRITE, 0, 0, *mmap_size, (void *)fixed_addr);
            if (addr == nullptr)
            {
                addr = MapViewOfFile(handle, FILE_MAP_WRITE, 0, 0, *mmap_size);
                SHM_WARN("MapViewOfFileEx failed, path: %s, fixed address(%#llx),real addr(%#llx)", path, fixed_addr, addr);
            }
            if (addr != nullptr)
            {
                g_mapping_handles[addr] = handle;
                return addr;
            }
        END:
            SHM_ERROR("CreateFileMapping failed, path: %s, error: %s.", path, strerror(GetLastError()));
            CloseHandle(handle);
            return nullptr;
        }

        void *ShmObjectAttach(const char *path, uintptr_t fixed_addr, size_t *real_size, size_t *mmap_size)
        {
            SYSTEM_INFO si;
            GetSystemInfo(&si);
            size_t pagesize = si.dwPageSize;

            size_t align = pagesize - 1;

            HANDLE handle = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, path);

            void *addr = nullptr;
            if (handle == nullptr)
            {
                SHM_ERROR("OpenFileMapping(%zu, %s) failed error:%s", *real_size, path, strerror(GetLastError()));
                return nullptr;
            }

            if (*mmap_size < *real_size)
            {
                SHM_ERROR("mmap size(%#" PRIx64 ") is not enough, revised to real size(%#" PRIx64 ").", *mmap_size, *real_size);
                *mmap_size = *real_size;
            }
            *mmap_size = PALIGN_UP(*mmap_size, pagesize);

            if (fixed_addr)
            {
                uintptr_t base_addr = fixed_addr;
                if (base_addr & align)
                {
                    base_addr = PALIGN_UP(fixed_addr, pagesize);
                    SHM_WARN("fixed address(%#" PRIx64 ") is not aligned to page size(%#" PRIx64 "), revised to %#" PRIx64 ".", base_addr, pagesize, fixed_addr);
                    fixed_addr = base_addr;
                }
            }
            // ������ӽ��������ַ��ռ�þͱ����ɣ���Ȼ�ָ���ַ����
            addr = MapViewOfFileEx(handle, FILE_MAP_WRITE, 0, 0, *mmap_size, (void *)fixed_addr);
            if (addr == nullptr)
            {
                SHM_ERROR("MapViewOfFileEx fixed address(%#" PRIx64 ") (%zu, %s) failed error:%s", fixed_addr, *real_size, path, strerror(GetLastError()));
                return nullptr;
            }
            if (addr != nullptr)
            {
                g_mapping_handles[addr] = handle;
                return addr;
            }

            SHM_ERROR("CreateFileMapping failed, path: %s, error: %s.", strerror(GetLastError()));
            CloseHandle(handle);
            return nullptr;
        }

        void ShmObjectDelete(void *addr, uintptr_t mmap_size, const char *path)
        {
            if (UnmapViewOfFile(addr) == 0)
            {
                SHM_ERROR("UnmapViewOfFile() failed, addr(%p), mmap size(%#" PRIx64 "), error: %s.", addr, mmap_size, strerror(GetLastError()));
            }
            auto itor = g_mapping_handles.find(addr);
            if (itor == g_mapping_handles.end())
            {
                SHM_ASSERT(false);
                return;
            }
            CloseHandle(itor->second);
            g_mapping_handles.erase(itor);
        }
        bool ShmObjectResize(const char *path, size_t *new_size)
        {
            return false;
        }

        bool GetAvailableMemorySize(size_t *const size)
        {
            /*HANDLE handle = GetCurrentProcess();
            if(handle == nullptr)
            {
                return false;
            }
            PROCESS_MEMORY_COUNTERS pmc;
            GetProcessMemoryInfo(handle, &pmc, sizeof(pmc));
            *size = pmc.PeakWorkingSetSize - pmc.WorkingSetSize;
            return true;*/
            return false;
        }
    }
}

#endif