﻿#include "vptr_manager.h"
#include "shm.h"
#include <memory>
#include <inttypes.h>
#define  VPTR_MAGIC 20230430

#define CONVERT_PTR_INXDE(ptr) (*(reinterpret_cast<uint32_t*>(ptr) - 1))
#define CONVERT_PTR_MAGIC(ptr) (*(reinterpret_cast<uint32_t*>(ptr) - 2))
#define CONVERT_RAW_PTR(ptr) ((void*)(reinterpret_cast<uint32_t*>(ptr) - 2))
namespace bg
{
    namespace detail
    {
        VptrManager* mgr;
        class HighOrderContainer;
        std::unique_ptr<HighOrderContainer> hoc;

        class HighOrderContainer
        {
        public:
            VptrObjectContainerBase* FindContainer(const bg::detail::TypeName& type_name)
            {
                auto itor = m_type_to_containers.find(type_name);
                if(itor != m_type_to_containers.end())
                {
                    return itor->second;
                }
                SHM_INFO("type(%s) not found!", type_name.c_str());
                return nullptr;
            }


            bool FixContainerVptr(VptrObjectContainerBase* fix_ptr)
            {
                const auto& type_name = fix_ptr->GetTypeName();
                auto itor = m_type_to_containers.find(type_name);
                if(itor == m_type_to_containers.end())
                {
                    SHM_ERROR("type(%s) not found!", type_name.c_str());
                    return false;
                }

                const uintptr_t* new_vptr = reinterpret_cast<const uintptr_t*>(reinterpret_cast<const char*>(itor->second));
                uintptr_t* old_vptr = reinterpret_cast<uintptr_t*>(reinterpret_cast<char*>(fix_ptr));
                if(*old_vptr != *new_vptr)
                {
                    *old_vptr = *new_vptr;
                }
                SHM_INFO("type(%s), old vptr(%p), new vptr(%p).", type_name.c_str(), old_vptr, new_vptr);
                return true;
            }

            VptrObjectContainerBase* NewShmContainer(const bg::detail::TypeName& type_name)
            {
                auto itor = m_type_to_containers.find(type_name);

                SHM_ASSERT_RETURN_NULL(itor != m_type_to_containers.end());

                return itor->second->ShmNew();
            }

            void RegisterContainer(VptrObjectContainerBase* obj)
            {
                auto& type_name = obj->GetTypeName();
                auto itor = m_type_to_containers.find(type_name);
                SHM_ASSERT_RETURN_VOID(itor == m_type_to_containers.end());

                m_type_to_containers[type_name] = obj;

            }

            template<typename Key, typename Value, typename Hash = std::hash<Key>, typename Pred = std::equal_to<Key>>
            using Container = std::unordered_map<Key, Value, Hash, Pred>;
            Container<TypeName, VptrObjectContainerBase*, TypeNameHasher> m_type_to_containers;
        };


        VptrObjectContainerBase::VptrObjectContainerBase(const TypeName& type, bool in_shm) :m_version(0LL), m_type(type)
        {
            if(in_shm)
            {
                std::vector<void*, bg::detail::ShmAllocator<void*> >* vector = (std::vector<void*, bg::detail::ShmAllocator<void*> > *)bg::ShmMalloc(sizeof(std::vector<void*, bg::detail::ShmAllocator<void*> >));
                if(vector)
                {
                    new (vector) std::vector<void*, bg::detail::ShmAllocator<void*> >();
                }
                m_objects = vector;
                this->m_version = bg::detail::ShmCurrentVersion();
            }
            else
            {
                if(!hoc)
                {
                    hoc = std::make_unique<HighOrderContainer>();
                }

                hoc->RegisterContainer(this);
            }
        }


        VptrObjectContainerBase::~VptrObjectContainerBase()
        {
            if(m_objects)
            {
                if(!m_objects->empty())
                {
                    SHM_ERROR("there are still %lu objects in container(%s)!", m_objects->size(), m_type.c_str());
                }
                m_objects->clear();
                bg::ShmFree(m_objects);
            }
            else if(m_base_types)
            {
                m_base_types->clear();
                delete m_base_types;
            }
        }

        void VptrObjectContainerBase::FixVptr()
        {
            if(m_version == bg::detail::ShmCurrentVersion())
            {
                SHM_DEBUG("container(%p: %s) already fixed, skip.", this, m_type.c_str());
            }
            else if(FixSelf())
            {
                FixObjects();
                this->m_version = bg::detail::ShmCurrentVersion();
            }
            else
            {
                SHM_INFO("container(%p: %s) seems to be defined in a dlopen() so, try fixing after dlopen() called.", this, m_type.c_str());
            }

        }

        const BaseTypeInfo* VptrObjectContainerBase::GetBaseTypeInfo(const TypeName& base_type)
        {
            auto container = hoc->FindContainer(m_type);
            if(!container)
            {
                return nullptr;
            }
            m_base_types = container->m_base_types;

            if(!m_base_types || m_base_types->empty())
            {
                return nullptr;
            }
            for(const auto& type : *m_base_types)
            {
                if(type.type == base_type)
                {
                    return &type;
                }
            }
            return nullptr;

        }

        void VptrObjectContainerBase::RecordObject(void* ptr)
        {
            if(ptr)
            {
                size_t size = m_objects->size();
                m_objects->push_back(ptr);
               CONVERT_PTR_INXDE(ptr) = static_cast<uint32_t>(size);
               CONVERT_PTR_MAGIC(ptr) = VPTR_MAGIC;
                SHM_INFO("object(%p) added, type: %s, index: %lu.", ptr, m_type.c_str(), size);
            }
        }

        void VptrObjectContainerBase::RemoveObject(void* ptr)
        {

            SHM_ASSERT_RETURN_VOID(ptr);
            uint32_t maric = CONVERT_PTR_MAGIC(ptr);
            if(maric != VPTR_MAGIC)
            {
                SHM_ERROR("invalid magic(%#x) for object(%p).", maric, ptr);
                return;
            }
            size_t size = m_objects->size();
            do
            {
                size_t index = CONVERT_PTR_INXDE(ptr);
              
                if(index >= size)
                {
                    SHM_ERROR("wrong index(%zu), size: %#" PRIx64", type: %s.", index, size, m_type.c_str());
                    break;
                }
                if((*m_objects)[index] == ptr)
                {
                    DoRemoveObject(index);
                    return;
                }
                SHM_ERROR("address mismatch(%p <=> %p), type: %s, index: %zu.", (*m_objects)[index], ptr, m_type.c_str(), index);

            } while(false);

            for(int i = 0; i < size; ++i)
            {
                if((*m_objects)[i] == ptr)
                {
                    SHM_INFO("found correct index: %d.", i);
                    DoRemoveObject(i);
                    return;
                }
            }
            SHM_ERROR("no stored object found with address(%p).", ptr);
            return;


        }

        bool VptrObjectContainerBase::FixSelf()
        {
            SHM_INFO(">>>>> fixing vptr of container(%p: %s)...", this, m_type.c_str());
            if(bg::detail::hoc && bg::detail::hoc->FixContainerVptr(this))
            {
                SHM_INFO("<<<<< vptr of container(%p: %s) fixed.", this, m_type.c_str());
                return true;
            }
            SHM_ERROR("<<<<< vptr of container(%p: %s) not fixed.", this, m_type.c_str());
            return false;
        }

        void VptrObjectContainerBase::DoRemoveObject(size_t index)
        {
            size_t size = m_objects->size();
            SHM_ASSERT_RETURN_VOID(index < m_objects->size());
            if(index == size - 1)
            {
                m_objects->pop_back();
            }
            else
            {
                auto ptr = (*m_objects)[size - 1];   
                CONVERT_PTR_INXDE(ptr) = static_cast<uint32_t>(index);
                (*m_objects)[index] = ptr;
                m_objects->pop_back();
            }
        }

        class VptrManager::Impl
        {
        public:
            template<typename Key, typename Value, typename Hash = std::hash<Key>, typename Pred = std::equal_to<Key>>
            using Container = BGSHM_API std::unordered_map<Key, Value, Hash, Pred, ShmAllocator<std::pair<const Key, Value>>>;
            Container<TypeName, VptrObjectContainerBase*, TypeNameHasher> m_type_to_containers;
        };

        bool VptrManager::Init()
        {
            if(!bg::detail::mgr)
            {
                bg::detail::mgr = (bg::detail::VptrManager*)bg::ShmGetSingleton<bg::detail::VptrManager>();
            }
            return bg::detail::mgr != nullptr;
        }

        void VptrManager::Fini()
        {
            if(bg::detail::mgr)
            {
                bg::ShmDeleteSingleton<bg::detail::VptrManager>();
                bg::detail::mgr = nullptr;
            }
        }

        VptrManager::VptrManager()
        {
                m_impl = (VptrManager::Impl*)bg::ShmMalloc(sizeof(VptrManager::Impl));
                SHM_ASSERT_RETURN_VOID(m_impl);

                new(m_impl) VptrManager::Impl();         
        }

        VptrManager::~VptrManager()
        {
            for(auto& item : mgr->m_impl->m_type_to_containers)
            {
                item.second->~VptrObjectContainerBase();
                bg::ShmFree(item.second);
            }
            mgr->m_impl->m_type_to_containers.clear();
            ShmFree(m_impl);
        }

        void VptrManager::FixVptr()
        {
            for(auto& item : mgr->m_impl->m_type_to_containers)
            {
                item.second->FixVptr();
            }
        }

        void* VptrManager::NewSpace(size_t bytes)
        {

            char* result = (char*)bg::ShmMalloc(bytes + sizeof(uint32_t) * 2);
            if(result)
            {
                result += 8;
            }
            return result;
        }

        void VptrManager::FreeSpace(void* ptr)
        {                       
            if(CONVERT_PTR_MAGIC(ptr) == VPTR_MAGIC)
            {
                bg::ShmFree(CONVERT_RAW_PTR(ptr));
            }
            else
            {
                SHM_ERROR("invalid magic(%p) at address(%p).", CONVERT_RAW_PTR(ptr), ptr);
            }
        }



        void* VptrManager::GetRealPtr(const TypeName& base_type, const TypeName& real_type, void* ptr)
        {
            if(base_type == real_type)
            {
                return ptr;
            }

            auto itor = bg::detail::mgr->m_impl->m_type_to_containers.find(real_type);

            SHM_ASSERT_RETURN_VALUE(itor != bg::detail::mgr->m_impl->m_type_to_containers.end(), ptr);

            auto info = itor->second->GetBaseTypeInfo(base_type);
            SHM_ASSERT_RETURN_VALUE(info, ptr);
            return &(((char*)ptr)[-info->offset]);
        }

        void VptrManager::RecordObject(const TypeName& type, void* ptr)
        {
            auto itor = bg::detail::mgr->m_impl->m_type_to_containers.find(type);
            if(itor == bg::detail::mgr->m_impl->m_type_to_containers.end())
            {
                VptrObjectContainerBase* base_ptr = hoc->NewShmContainer(type);
                SHM_ASSERT_RETURN_VOID(base_ptr && base_ptr->GetTypeName() == type);
                auto ret = bg::detail::mgr->m_impl->m_type_to_containers.insert(std::make_pair(type, base_ptr));
                SHM_ASSERT_RETURN_VOID(ret.second && ret.first != mgr->m_impl->m_type_to_containers.end());
                itor = ret.first;
            }


            itor->second->RecordObject(ptr);
            SHM_DEBUG("container(%p) for type(%s) created.", itor->second, type.c_str());
        }

        void VptrManager::RemoveObject(const TypeName& type, void* ptr)
        {
            auto itor = bg::detail::mgr->m_impl->m_type_to_containers.find(type);

            SHM_ASSERT_RETURN_VOID(itor != bg::detail::mgr->m_impl->m_type_to_containers.end());
            itor->second->RemoveObject(ptr);
        }


    }
}
